﻿using System;
using System.Collections;
using System.Linq;
using UnityEditor;
using UnityEngine;

/// <summary>
/// <see cref="Fact.Id"/>/ <c>MonoBehaviour</c> wrapper to be attached to <see cref="Fact.WorldRepresentation"/>
/// </summary>
public class FactObject3D : FactObject
{
    public new Renderer[] renderer;

    #region Unity Serialization
    protected override void _OnBeforeSerialize()
    {
#if UNITY_EDITOR // not working
        if (EditorApplication.isPlaying)
            return;

        foreach (Collider collider in transform.GetComponentsInChildren<Collider>(includeInactive: true))
        {
            if (!collider.gameObject.GetComponent<FactObject>())
                Debug.LogWarningFormat(
                    "{0}: {1} under {2} {3} has missing {2}!\n" +
                    "This is needed to select a {2}!",
                    nameof(Collider), collider, nameof(FactObject3D), this);
            //try
            //{
            //    collider.gameObject.AddComponent<FactObject3D>();
            //}
            //catch (Exception) { }
        }
#endif
    }

    protected override void _OnAfterDeserialize() { }
    #endregion Unity Serialization

    protected override void _Awake()
    {
        base._Awake();

        MaterialsToChange ??= renderer?.SelectMany(r => r.materials).ToArray();
    }

    protected override void _CustomReLabel()
    {
        switch (Fact) // for highly customized labels
        {
            case TestFact testFact:
                //FactText[0].text = testFact.Label;
                break;
            default:
                break;
        }
    }

    public void CascadeForMeAndChildren(Action<FactObject3D> func)
    {
        func(this);
        foreach (FactObject3D fo in AllChildren.OfType<FactObject3D>())
            func(fo);
    }

    public void ForAllRenderer(Action<Renderer> func)
    {
        foreach (Renderer ren in renderer)
            func(ren);
    }

    public void CoroutineCascadeForMeAndChildrenAllRenderer(Func<FactObject3D, Renderer, IEnumerator> func)
    {
        this.StopAllCoroutines();

        CascadeForMeAndChildren((FactObject3D fo) =>
            fo.ForAllRenderer((Renderer ren) =>
                this.StartCoroutine(func(fo, ren))
        ));
    }
}
