using Newtonsoft.Json;
using REST_JSON_API;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Runtime.CompilerServices;
using Unity.Mathematics;
using UnityEngine;

/// <summary>
/// 
/// </summary>
/// <seealso cref="SOMDoc.MakeDeepList(SOMDoc[], SOMDoc)"/>
/// <remarks>Needed to refere to lists serverside</remarks>
public class ListFact : FactWrappedCRTP<ListFact>
{
    protected override object GetCompiledValue() => CompileValue(Defines());

    public string[] lids = new string[0];

    public SOMDoc[] payload = new SOMDoc[0];

    public OMA ListType;

    public ListFact() : base() { }

    public ListFact(string[] lids, SOMDoc[] payload, SOMDoc ElementType) : base()
    {
        Init(lids, payload, ElementType);
    }

    public ListFact(string[] lids, SOMDoc[] payload, SOMDoc ElementType, SOMDoc _ServerDefinition) : base()
    {
        Init(lids, payload, ElementType);
        this.ServerDefinition = _ServerDefinition;
    }

    private void Init(string[] lids, SOMDoc[] payload, SOMDoc ElementType)
    {
        this.lids = lids ?? new string[0];

        this.payload = new SOMDoc[math.max(this.lids.Length, payload?.Length ?? 0)];
        payload?.CopyTo(this.payload, 0);

        if (!HasDependentFacts)
        {
            if (payload.Any(p => p == null))
                throw new ArgumentException(nameof(payload) + "must not include null elements iff " + nameof(lids) + " is all nulls or empty!");
        }
        else
        {
            for (int i = 0; i < this.lids.Length; i++)
                if (this.lids[i] != null)
                    this.payload[i] = new OMS(this.lids[i]);
                else
                if (this.payload[i] == null)
                    throw new ArgumentException(nameof(lids) + " and " + nameof(payload) + " have to be complementary not null!");
        }

        if (ElementType == null)
        {
            if (this.payload.Length == 0)
                throw new ArgumentException(nameof(ElementType) + " must not be null if ListFact is empty List!");

            ElementType = payload[0].SOMDocType();
        }

        this.ListType = new OMA(
            new OMS(MMTConstants.ListType),
            new[] { ElementType }
        );
    }

    protected override Fact _ReInitializeMe(Dictionary<string, string> old_to_new)
        => new ListFact(
            lids.Select(id => id == null ? null : old_to_new[id]).ToArray(),
            payload.Select(somdoc => somdoc?.MapURIs(old_to_new)).ToArray(),
            ListType.arguments[0]);

    public static new IEnumerable parseFact(List<Fact> ret, MMTFact fact)
    {
        if (fact is not MMTGeneralFact MMTSymbol)
            yield break;

        if (MMTSymbol.type is not OMA listOMA
         || listOMA.arguments[0] is not OMA typeOMA)
        {
            // try to salvage
            ret.AddRange(DynamicListFact.MMTFactory(null,
                MMTSymbol.defines, MMTSymbol.type,
                MMTSymbol.@ref
            ));
            yield break;
        }

        List<SOMDoc> payload;

        if (MMTSymbol.defines is OMA OMAList
         && OMAList.applicant is OMS OMSList
         && OMSList.uri == MMTConstants.ListApplicant)
        {
            payload = OMAList.arguments.ToList();
        }
        else
        {
            payload = new();
            SOMDoc next_element = MMTSymbol.defines;
            int depth = 0;
            while (true)
            {
                ++depth;

                if (next_element is not OMA current_element
                 || current_element.applicant is not OMS OMSapp)
                { // neither ListLiteral nor ListEnd
                    throw new ArgumentException("Malformed List-Format");
                }

                if (OMSapp.uri != MMTConstants.ListLiteral)
                {
                    if (OMSapp.uri != MMTConstants.ListEnd)
                    { // neither ListLiteral nor ListEnd
                        throw new ArgumentException("Malformed List-Format");
                    }

                    break;
                }

                if (current_element.arguments.Length < 2) // Maybe empty list?
                    throw new ArgumentException("Unrecognized List-Format");

                payload.Add(current_element.arguments[1]);
                next_element = current_element.arguments[0];
            }
        }

        ret.Add(new ListFact(
            payload.Select(p => (p as OMS)?.uri).ToArray(),
            payload.ToArray(),
            listOMA.arguments[0],
            fact.@ref));
    }

    public override MMTFact MakeMMTDeclaration()
        => new MMTGeneralFact(_LastLabel, GetMMTType(), Defines());

    public override SOMDoc Defines()
        => SOMDoc.MakeShallowList(payload);

    public override SOMDoc GetMMTType()
        => ListType;

    protected override string[] GetDependentFactIds()
        => lids.Where(lid => lid != null).ToArray();

    protected override void RecalculateTransform() { }

    protected override bool EquivalentWrapped(ListFact f1, ListFact f2)
        => f1.ListType.Equivalent(f2.ListType)
        && f1.payload.Length == f2.payload.Length
        && DependentFactsEquivalent(f1, f2)
        && f1.payload
            .Zip(f2.payload, (p1, p2) => SOMDoc.Equivalent(p1, p2))
            .All(b => b);
}

public class DynamicListFact : FactWrappedCRTP<DynamicListFact>, IUnpackable
{
    protected override object GetCompiledValue() => CompileValue(indirect_payload);

    [JsonIgnore]
    public List<object> payload;

    public SOMDoc indirect_payload;

    public SOMDoc ListType;

    public DynamicListFact() : base() { }

    //public DynamicListFact(List<object> payload, SOMDoc indirect_payload, SOMDoc ListType) : base()
    //{
    //    this.payload = payload;
    //    this.indirect_payload = indirect_payload;
    //    this.ListType = ListType;
    //}

    public static List<Fact> MMTFactory(List<dynamic> payload, SOMDoc indirect_payload, SOMDoc ListType, SOMDoc _ServerDefinition = null)
    {
        if (payload == null
         && indirect_payload == null)
            return new();

        Expression CompileBase = null;
        object CompiledValue = payload;
        if (payload == null)
        {
            Func<object[], object[]> builder =
                indirect_payload.PartialInvokeCastingLambdaExpression(out CompileBase, out Type[] signature);

            if (signature.Length != 1)
                return new();

            CompiledValue = builder(new object[0])[0];

            if (CompiledValue is List<dynamic> list_result)
                payload = list_result;
            else
            if (CompiledValue is IEnumerable<dynamic> enum_result)
                payload = enum_result.ToList();
            else
            if (CompiledValue is IEnumerable<float> enum_float) // special case for boxed/struct values
                payload = enum_float.Select(f => (dynamic)f).ToList();
            else
            if (signature[0].IsArray)
                throw new NotSupportedException();
            else
                return new();
        }

        indirect_payload ??= SOMDoc.SOMDocObject(payload);
        ListType ??= SOMDoc.SOMDocType(CompiledValue?.GetType() ?? payload.GetType());

        List<Fact> ret = new() {
            new DynamicListFact()
        {
            CompiledValue = CompiledValue,
            payload = payload.Select(e => (object) e).ToList(),
            indirect_payload = indirect_payload,
            ListType = ListType,
        }};

        if (CompileBase != null)
            (ret[0] as DynamicListFact).CompileBase = CompileBase;

        if (_ServerDefinition != null)
            (ret[0] as DynamicListFact).ServerDefinition = _ServerDefinition;

        return ret;
    }

    public override MMTFact MakeMMTDeclaration()
        => Defines() == null
        ? throw new NotSupportedException(nameof(indirect_payload) + " must not be null")
        : new MMTGeneralFact(_LastLabel, ListType, Defines());

    public override SOMDoc Defines()
    {
        try
        {
            return SOMDoc.SOMDocObject(CompiledValue);
        }
        catch (Exception)
        {
            return indirect_payload;
        }
    }

    public new static IEnumerator parseFact(List<Fact> ret, MMTFact fact)
    {
        if (fact is not MMTGeneralFact MMTSymbol)
            yield break;

        List<Fact> orig = MMTFactory(null,
            MMTSymbol.defines, MMTSymbol.type,
            MMTSymbol.@ref
        );

        ret.AddRange(orig);
        foreach (Fact Fact in orig)
            if (Fact is DynamicListFact pack)
                yield return pack.UnpackMe(ret, true); //TODO: set to false, once Server can index
    }

    public SOMDoc SOMDocIndexer(int i)
        => new OMA(
            new OMS(MMTConstants.IndexList),
            new[] {
                indirect_payload,
                new OMLIT<int>(i)
        });

    protected override bool EquivalentWrapped(DynamicListFact f1, DynamicListFact f2)
        => f1.payload != null && f2.payload != null
        && f1.payload.Count == f2.payload.Count
        && f1.ListType.Equivalent(f2.ListType)
        && f1.payload.SequenceEqual(f2.payload, ApproximationComparer.Instance);

    protected override string[] GetDependentFactIds()
        => indirect_payload.GetDependentFactIds();

    protected override void RecalculateTransform() { }

    protected override Fact _ReInitializeMe(Dictionary<string, string> old_to_new)
        => MMTFactory(payload, indirect_payload?.MapURIs(old_to_new), ListType) //, ServerDefinition?.MapURIs(old_to_new))
            .FirstOrDefault();

    public IEnumerator UnpackMe(List<Fact> ret, bool BypassServer)
    {
        for (int i = 0; i < payload.Count; i++)
        {
            MakeFact(ret, payload[i], SOMDocIndexer(i), BypassServer);
            yield return null;
        }
    }
}

/// <summary>
/// 
/// </summary>
/// <seealso cref="SOMDoc.MakeTupel(SOMDoc[])"/>
/// <remarks>Needed to refere to tupels serverside</remarks>
public class TupleFact : FactWrappedCRTP<TupleFact>
{
    protected override object GetCompiledValue() => CompileValue(Defines());

    public string[] lids = new string[0];

    public SOMDoc[] payload = new SOMDoc[0];


    public TupleFact() : base() { }

    public TupleFact(string[] lids, SOMDoc[] payload) : base()
    {
        Init(lids, payload);
    }

    public TupleFact(string[] lids, SOMDoc[] payload, SOMDoc _ServerDefinition) : base()
    {
        Init(lids, payload);
        this.ServerDefinition = _ServerDefinition;
    }

    private void Init(string[] lids, SOMDoc[] payload)
    {
        this.lids = lids ?? new string[0];
        if (!HasDependentFacts)
        {
            if (payload.Any(p => p == null))
                throw new ArgumentException(nameof(payload) + "must not include null elements iff " + nameof(lids) + " is all nulls or empty!");

            this.payload = payload;
        }
        else
        {
            this.payload = new SOMDoc[math.max(lids.Length, payload.Length)];

            for (int i = 0; i < this.payload.Length; i++)
                if (lids[i] != null)
                    payload[i] = new OMS(lids[i]);
                else
                if (lids[i] == null && payload[i] == null)
                    throw new ArgumentException(nameof(lids) + " and " + nameof(payload) + " have to be complementary-not-null!");
        }
    }

    protected override Fact _ReInitializeMe(Dictionary<string, string> old_to_new)
        => new TupleFact(
            lids.Select(id => id == null ? null : old_to_new[id]).ToArray(),
            payload.Select(somdoc => somdoc == null ? null : somdoc.MapURIs(old_to_new)).ToArray());

    public static new IEnumerable parseFact(List<Fact> ret, MMTFact fact)
    {
        if (fact is not MMTGeneralFact MMTSymbol
         || MMTSymbol.defines is not OMA defineOMA)
        {
            yield break;
        }

        string[] lids = defineOMA.arguments
            .Select(oms => (oms as OMS)?.uri)
            .ToArray();

        ret.Add(new TupleFact(lids, defineOMA.arguments, fact.@ref));
    }

    public override MMTFact MakeMMTDeclaration()
    {
        SOMDoc Tupel = Defines();
        return new MMTGeneralFact(_LastLabel, Tupel.SOMDocType(), Tupel);
    }

    public override SOMDoc Defines()
        => SOMDoc.MakeTupel(payload);

    protected override string[] GetDependentFactIds()
        => lids.Where(lid => lid != null).ToArray();

    protected override void RecalculateTransform() { }

    protected override bool EquivalentWrapped(TupleFact f1, TupleFact f2)
        => f1.payload.Length == f2.payload.Length
        && DependentFactsEquivalent(f1, f2)
        && f1.payload
            .Zip(f2.payload, (p1, p2) => (p1 is OMS && p2 is OMS) || SOMDoc.Equivalent(p1, p2))
            .All(b => b);
}

public class DynamicTupleFact : FactWrappedCRTP<DynamicTupleFact>, IUnpackable
{
    protected override object GetCompiledValue() => payload;

    [JsonIgnore]
    object payload = null;

    int count = 0;

    [JsonIgnore]
    public Type[] Signature => _Signature ??= payload.GetType().GetGenericArguments();

    private Type[] _Signature;

    public SOMDoc indirect_payload;

    public SOMDoc TupleType;

    public DynamicTupleFact() : base() { }

    public static List<Fact> MMTFactory(object payload, SOMDoc indirect_payload, SOMDoc TupleType, SOMDoc _ServerDefinition = null)
    {
        if (payload == null
         && indirect_payload == null)
            return new();

        if (payload == null)
        {
            Func<object[], object[]> builder =
                indirect_payload.PartialInvokeCastingLambdaExpression(out _, out Type[] signature);

            if (signature.Length != 1)
                return new();

            payload = builder(new object[0])[0];
        }

        if (!TupleExtensions.IsTupleType(payload.GetType(), out int count))
            return new();
        Type[] Signature = payload.GetType().GetGenericArguments();
        List<Fact> ret = new();

        // Isomorphismen
        if (payload is Tuple<float, float, float> point)
        {
            Vector3 vec = new(point.Item1, point.Item2, point.Item3);
            ret.Add(new PointFact(vec, vec));
            return ret;
        }
        //if(payload is Tuple<Tuple<float, float>, Func<float, TIn>, Func<TIn?[..],TOut>> func)
        if (Signature.Length == 3
         && Signature[0].Equals(typeof(Tuple<float, float>))
         && FuncExtensions.IsFuncType(Signature[1], out int callerCount)
         && FuncExtensions.IsFuncType(Signature[2], out int calledCount)
         && callerCount == 2 && calledCount >= 2
         && Signature[1].GetGenericArguments()[0].Equals(typeof(float)))
        {
            Type callerType = Signature[1].GetGenericArguments()[1];
            Type[] calledType = Signature[2].GetGenericArguments();

            if ((calledCount == 2
                && callerType.Equals(calledType[0])
             ) || (callerType.IsGenericType
                && calledCount - 1 == callerType.GetGenericArguments().Length
                && callerType.GetGenericArguments()
                    .Zip(calledType[..^1], (a, b) => a.Equals(b))
                    .All(b => b)
            ))
            {
                Tuple<float, float> domain = (Tuple<float, float>)((ITuple)payload)[0];
                ret.Add(new FunctionFact(SOMDocIndexer(1, 3, indirect_payload)));
                ret.Add(new FunctionFact(SOMDocIndexer(2, 3, indirect_payload)));
                ret.Add(new FunctionCallFact((FunctionFact)ret[1], (FunctionFact)ret[0], (domain.Item1, domain.Item2)));
                return ret;
            }
        }

        return new(); // Don't care about Tuple without semantic meaning

        ////TODO? indirect_payload ??= SOMDoc.FromObject(payload);
        //TupleType ??= SOMDoc.SOMDocType(payload.GetType());

        //ret.Add(new DynamicTupleFact(organizer)
        //{
        //    payload = payload,
        //    indirect_payload = indirect_payload,
        //    TupleType = TupleType,
        //    count = count,
        ////    ServerDefinition = _ServerDefinition,
        //});

        //if (_ServerDefinition != null)
        //    ret[0].ServerDefinition = _ServerDefinition;

        //return ret;
    }

    public override MMTFact MakeMMTDeclaration()
        => Defines() == null
        ? throw new NotSupportedException(nameof(indirect_payload) + " must not be null")
        : new MMTGeneralFact(_LastLabel, GetMMTType(), Defines());

    public override SOMDoc Defines()
    {
        try
        {
            return SOMDoc.SOMDocObject(CompiledValue);
        }
        catch (NotSupportedException)
        {
            return indirect_payload;
        }
    }

    public override SOMDoc GetMMTType()
        => TupleType;

    public new static IEnumerator parseFact(List<Fact> ret, MMTFact fact)
    {
        if (fact is not MMTGeneralFact MMTSymbol)
            yield break;

        ret.AddRange(
            MMTFactory(null,
                MMTSymbol.defines, MMTSymbol.type,
                MMTSymbol.@ref
        ));
    }

    public SOMDoc SOMDocIndexer(int i)
        => SOMDocIndexer(i, count, indirect_payload);

    public static SOMDoc SOMDocIndexer(int i, int count, SOMDoc indirect_payload)
    {
        if (i >= count && i < 0)
            throw new ArgumentOutOfRangeException($"{nameof(i)} must not be negative or exceed {count}");

        SOMDoc ret = indirect_payload;

        for (int j = 0; j < i; j++)
            ret = new OMA(
                new OMS(MMTConstants.ProjR),
                new[] { ret }
            );

        return count - 1 - i > 0
            ? new OMA(
                new OMS(MMTConstants.ProjL),
                new[] { ret })
            : ret;
    }

    protected override bool EquivalentWrapped(DynamicTupleFact f1, DynamicTupleFact f2)
    {
        if (f1.count != f2.count
        || !f1.TupleType.Equivalent(f2.TupleType)
        || f1.payload == null || f2.payload == null)
            return false;

        for (int i = 0; i < f1.count; i++)
            if (!ApproximationComparer.Instance
                .Equals(((ITuple)f1.payload)[i], ((ITuple)f2.payload)[i]))
                return false;

        return true;
    }

    protected override string[] GetDependentFactIds()
        => indirect_payload.GetDependentFactIds();

    protected override void RecalculateTransform() { }

    protected override Fact _ReInitializeMe(Dictionary<string, string> old_to_new)
        => MMTFactory(payload, indirect_payload?.MapURIs(old_to_new), TupleType)?[0];

    public IEnumerator UnpackMe(List<Fact> ret, bool BypassServer)
    {
        for (int i = 0; i < count; i++)
        {
            MakeFact(ret, ((ITuple)payload)[i], SOMDocIndexer(i), BypassServer);
            yield return null;
        }
    }
}

public class RealLitFact : FactWrappedCRTP<RealLitFact>
{
    protected override object GetCompiledValue() => value;

    public float value;

    public RealLitFact() : base() { }

    public RealLitFact(float value) : base()
    {
        this.value = value;
    }

    public RealLitFact(float value, SOMDoc _ServerDefinition) : base()
    {
        this.value = value;
        this.ServerDefinition = _ServerDefinition;
    }

    public override MMTFact MakeMMTDeclaration()
    {
        SOMDoc df = Defines();

        return new MMTGeneralFact(
                _LastLabel,
                df.SOMDocType(),
                df
            );
    }

    public override SOMDoc Defines()
        => new OMLIT<float>(value);

    public override SOMDoc GetMMTType()
        => new OMS(MMTConstants.RealLit);

    protected override bool EquivalentWrapped(RealLitFact f1, RealLitFact f2)
        => Mathf.Approximately(f1.value, f2.value);

    protected override string[] GetDependentFactIds()
        => new string[0];

    protected override void RecalculateTransform() { }

    protected override Fact _ReInitializeMe(Dictionary<string, string> old_to_new)
        => new RealLitFact(value);
}

public class GeneralFact : FactWrappedCRTP<GeneralFact>
{
    protected override object GetCompiledValue() => CompileValue(Defines());

    public MMTFact Raw { get; private set; }

    //[JsonIgnore]
    //public LambdaExpression GeneratedExpression { get; private set; }

    public GeneralFact() : base() { }
    public GeneralFact(MMTFact Raw) : base()
    {
        this.Raw = Raw;
    }
    public GeneralFact(MMTFact Raw, SOMDoc ServerDefinition) : base()
    {
        this.Raw = Raw;
        this.ServerDefinition = ServerDefinition;
    }

    public new static IEnumerator parseFact(List<Fact> ret, MMTFact fact)
    {
        ret.Add(new GeneralFact(fact, fact.@ref));
        yield break;
    }

    public override MMTFact MakeMMTDeclaration()
        => Raw;

    public override SOMDoc Defines()
        => Raw.GetDefines();

    public override SOMDoc GetMMTType()
        => Raw.GetMMTType() is OMS omsType && omsType.uri == MMTConstants.TypeType
         ? Raw.GetDefines()
         : Raw.GetMMTType();

    protected override bool EquivalentWrapped(GeneralFact f1, GeneralFact f2)
        => MMTFact.Equivalent(f1.Raw, f2.Raw);

    protected override string[] GetDependentFactIds()
        => Raw.GetDependentFactIds();

    protected override void RecalculateTransform() { }

    protected override Fact _ReInitializeMe(Dictionary<string, string> old_to_new)
        => new GeneralFact(Raw.MapURIs(old_to_new));
}